"""
Module related to abstract attribution metric
"""

from abc import ABC, abstractmethod

import tensorflow as tf
import numpy as np

from ..commons import Tasks, numpy_sanitize, get_inference_function
from ..types import Callable, Optional, Union, OperatorSignature


class BaseAttributionMetric(ABC):
    """
    Base class for Attribution Metric.

    Parameters
    ----------
    model
        Model used for computing explanations.
    inputs
        Input samples to be explained.
    targets
        One-hot encoded labels or regression target (e.g {+1, -1}), one for each sample.
    batch_size
        Number of samples to evaluate at once, if None compute all at once.
    activation
        A string that belongs to [None, 'sigmoid', 'softmax']. Specify if we should add
        an activation layer once the model has been called. It is useful, for instance
        if you want to measure a 'drop of probability' by adding a sigmoid or softmax
        after getting your logits. If None does not add a layer to your model.
    """

    def __init__(self,
                 model: Callable,
                 inputs: Union[tf.data.Dataset, tf.Tensor, np.ndarray],
                 targets: Optional[Union[tf.Tensor, np.ndarray]] = None,
                 batch_size: Optional[int] = 64,
                 activation: Optional[str] = None):
        if activation is None:
            self.model = model
        else:
            assert activation in ['sigmoid', 'softmax'], \
            "activation must be in ['sigmoid', 'softmax']"
            if activation == 'sigmoid':
                self.model = lambda x: tf.nn.sigmoid(model(x))
            else:
                self.model = lambda x: tf.nn.softmax(model(x), axis=-1)
        self.inputs, self.targets = numpy_sanitize(inputs, targets)
        self.batch_size = batch_size


class ExplainerMetric(BaseAttributionMetric, ABC):
    """
    Base class for Attribution Metric that require explainer.

    Parameters
    ----------
    model
        Model used for computing explanations.
    inputs
        Input samples to be explained.
    targets
        One-hot encoded labels or regression target (e.g {+1, -1}), one for each sample.
    batch_size
        Number of samples to evaluate at once, if None compute all at once.
    activation
        A string that belongs to [None, 'sigmoid', 'softmax']. Specify if we should add
        an activation layer once the model has been called. It is useful, for instance
        if you want to measure a 'drop of probability' by adding a sigmoid or softmax
        after getting your logits. If None does not add a layer to your model.
    """

    @abstractmethod
    def evaluate(self,
                 explainer: Callable) -> float:
        """
        Compute the score of the given explainer.

        Parameters
        ----------
        explainer
            Explainer to call to get explanation for an input and a label.

        Returns
        -------
        score
            Score of the explainer on the inputs.
        """
        raise NotImplementedError()

    def __call__(self,
                 explainer: Callable) -> float:
        """Evaluate alias"""
        return self.evaluate(explainer)


class ExplanationMetric(BaseAttributionMetric, ABC):
    """
    Base class for Attribution Metric that require explanations.

    Parameters
    ----------
    model
        Model used for computing explanations.
    inputs
        Input samples to be explained.
    targets
        One-hot encoded labels or regression target (e.g {+1, -1}), one for each sample.
    batch_size
        Number of samples to evaluate at once, if None compute all at once.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].
    activation
        A string that belongs to [None, 'sigmoid', 'softmax']. Specify if we should add
        an activation layer once the model has been called. It is useful, for instance
        if you want to measure a 'drop of probability' by adding a sigmoid or softmax
        after getting your logits. If None does not add a layer to your model.
    """
    def __init__(self,
                 model: tf.keras.Model,
                 inputs: Union[tf.data.Dataset, tf.Tensor, np.ndarray],
                 targets: Optional[Union[tf.Tensor, np.ndarray]] = None,
                 batch_size: Optional[int] = 64,
                 operator: Optional[Union[Tasks, str, OperatorSignature]] = None,
                 activation: Optional[str] = None,
                 ):
        # pylint: disable=R0913
        super().__init__(model, inputs, targets, batch_size, activation)

        # define the inference function according to the model type
        self.inference_function, self.batch_inference_function = \
            get_inference_function(model, operator)

    @abstractmethod
    def evaluate(self,
                 explanations: Union[tf.Tensor, np.array]) -> float:
        """
        Compute the score of the given explanations.

        Parameters
        ----------
        explanations
            Explanation for the inputs, labels to evaluate.

        Returns
        -------
        score
            Score of the explanations.
        """
        raise NotImplementedError()

    def __call__(self,
                 explanations: Union[tf.Tensor, np.array]) -> float:
        """Evaluate alias"""
        return self.evaluate(explanations)



import os
import random

import numpy as np
import tensorflow as tf

from xplique.attributions import Rise
from ..utils import generate_data, generate_model, almost_equal


def test_output_shape():
    """The output size (h, w) must be the same as the input"""

    input_shapes = [(28, 28, 1), (32, 32, 3), (30, 50, 3)]
    nb_labels = 10

    for input_shape in input_shapes:
        x, y = generate_data(input_shape, nb_labels, 10)
        model = generate_model(input_shape, nb_labels)

        method = Rise(model, nb_samples=100)
        rise_maps = method.explain(x, y)
        assert x.shape[:-1] == rise_maps.shape

        method = Rise(model, nb_samples=100, grid_size=(3, 5))
        rise_maps = method.explain(x, y)
        assert x.shape[:-1] == rise_maps.shape


def reset_random_seed(seed: int = 0):
   os.environ['PYTHONHASHSEED'] = str(seed)
   tf.random.set_seed(seed)
   np.random.seed(seed)
   random.seed(seed)


def test_polymorphic_parameters():
    """Ensure we could pass tuple or int to define grid_size parameter when inputs are images"""
    grid_size = 3
    nb_samples = 10
    preservation_probability = 0.5

    input_shapes = [(28, 28, 1), (32, 32, 3)]
    nb_labels = 10

    for input_shape in input_shapes:
        features, targets = generate_data(input_shape, nb_labels, 20)
        model = generate_model(input_shape, nb_labels)

        reset_random_seed(0)
        rise_int = Rise(model, grid_size=grid_size, nb_samples=nb_samples)
        reset_random_seed(0)
        rise_tuple = Rise(model, grid_size=(grid_size, grid_size), nb_samples=nb_samples)

        reset_random_seed(0)
        mask_int = tf.cast(
            rise_int._get_masks(nb_samples, grid_size, preservation_probability), 
            tf.int32)
            
        reset_random_seed(0)
        mask_tuple = tf.cast(
            rise_tuple._get_masks(nb_samples, grid_size, preservation_probability), 
            tf.int32)

        rise_int(features, targets)
        rise_tuple(features, targets)

        assert almost_equal(mask_int, mask_tuple)



import matplotlib
import numpy as np
import os

from xplique.plots.metrics import barplot, fidelity_curves


def test_bar_plot():
    # test metrics barplot
    cmap = matplotlib.cm.get_cmap("Set3")
    methods_colors = {"method_" + str(i): cmap(i / 9) for i in range(10)}

    scores = {}
    for metric in ["metric_1", "metric_2"]:
        scores[metric] = {
            "method_" + str(i): np.random.random_sample()
            for i in range(10)
        }

    filepath = "tests/plots/barplot_test.png"
    barplot(
        scores,
        sort_metric="metric_1",
        ascending=True,
        title="test bar-plot",
        filepath=filepath,
        methods_colors=methods_colors
    )

    assert os.path.exists(filepath)
    os.remove(filepath)


def test_curves():
    # test fidelity metric curves plot
    cmap = matplotlib.cm.get_cmap("Set3")
    methods_colors = {"method_" + str(i): cmap(i / 9) for i in range(10)}

    steps = np.linspace(0, 100, num=11)
    detailed_scores = {
        "method_" + str(i):
            {step: val for (step, val) in
             zip(steps, np.linspace(*np.random.randint(0, 100, 2), num=11))}
        for i in range(10)
    }

    filepath = "tests/plots/fidelity_curves_test.png"
    fidelity_curves(
        detailed_scores,
        title="test curves",
        filepath=filepath,
        methods_colors=methods_colors
    )

    assert os.path.exists(filepath)
    os.remove(filepath)



"""
Override relu gradients policy
"""

import tensorflow as tf
from tensorflow.keras.models import clone_model # pylint: disable=E0611

from ..types import Tuple, Callable, Union, Optional


def guided_relu_policy(max_value: Optional[float] = None,
                       threshold: float = 0.0) -> Callable:
    """
    Generate a guided relu activation function.
    Some models have relu with different threshold and plateau settings than a classic relu,
    it is important to preserve these settings when changing activations so that the forward
    remains intact.

    Parameters
    ----------
    max_value
        If specified, the maximum value for the ReLU.
    threshold
        If specified, the threshold for the ReLU.

    Returns
    -------
    guided_relu
        A guided relu activation function.
    """
    relu = tf.keras.layers.ReLU(max_value=max_value, threshold=threshold)

    @tf.custom_gradient
    def guided_relu(inputs: tf.Tensor) -> Tuple[tf.Tensor, Callable]:
        """
        Guided relu activation function.
        Act like a relu during forward pass, but allows only positive gradients with positive
        activation to pass through during backprop.

        Parameters
        ----------
        inputs
            Input tensor

        Returns
        -------
        output
            Tensor, output or relu transformation.
        grad_func
            Gradient function for guided relu.
        """

        def grad_func(grads):
            gate_activation = tf.cast(inputs > 0.0, tf.float32)
            return tf.nn.relu(grads) * gate_activation

        return relu(inputs), grad_func

    return guided_relu


def deconv_relu_policy(max_value: Optional[float] = None,
                       threshold: float = 0.0) -> Callable:
    """
    Generate a deconv relu activation function.
    Some models have relu with different threshold and plateau settings than a classic relu,
    it is important to preserve these settings when changing activations so that the forward
    remains intact.

    Parameters
    ----------
    max_value
        If specified, the maximum value for the ReLU.
    threshold
        If specified, the threshold for the ReLU.

    Returns
    -------
    deconv_relu
        A deconv relu activation function.
    """
    relu = tf.keras.layers.ReLU(max_value=max_value, threshold=threshold)

    @tf.custom_gradient
    def deconv_relu(inputs: tf.Tensor) -> Tuple[tf.Tensor, Callable]:
        """
        DeconvNet activation function.
        Act like a relu during forward pass, but allows only positive gradients to pass through
        during backprop.

        Parameters
        ----------
        inputs
            Input tensor

        Returns
        -------
        output
            Tensor, output or relu transformation.
        grad_func
            Gradient function for DeconvNet relu.
        """

        def grad_func(grads):
            return tf.nn.relu(grads)

        return relu(inputs), grad_func

    return deconv_relu


def open_relu_policy(max_value: Optional[float] = None,
                     threshold: float = 0.0) -> Callable:
    """
    Generate a relu activation function which allows gradients to pass.

    Parameters
    ----------
    max_value
        If specified, the maximum value for the ReLU.
    threshold
        If specified, the threshold for the ReLU.

    Returns
    -------
    open_relu
        A relu which allows all gradients to pass.
    """
    relu = tf.keras.layers.ReLU(max_value=max_value, threshold=threshold)

    @tf.custom_gradient
    def open_relu(inputs: tf.Tensor) -> Tuple[tf.Tensor, Callable]:
        """
        OpenRelu activation function.
        Act like a relu during forward pass, but allows all gradients to pass through
        during backprop.

        Parameters
        ----------
        inputs
            Input tensor

        Returns
        -------
        output
            Tensor, output or relu transformation.
        grad_func
            Gradient function for OpenRelu.
        """

        def grad_func(grads):
            return grads

        return relu(inputs), grad_func

    return open_relu


def is_relu(layer: tf.keras.layers.Layer) -> bool:
    """
    Check if a layer is a ReLU layer

    Parameters
    ----------
    layer
        Layer to check.

    Returns
    -------
    is_relu : bool
        True if the layer is a relu activation.
    """
    return isinstance(layer, tf.keras.layers.ReLU)


def has_relu_activation(layer: tf.keras.layers.Layer) -> bool:
    """
    Check if a layer has a ReLU activation.

    Parameters
    ----------
    layer
        Layer to check.

    Returns
    -------
    has_relu
        True if the layer has a relu activation.
    """
    if not hasattr(layer, 'activation'):
        return False
    return layer.activation in [tf.nn.relu, tf.keras.activations.relu]


def override_relu_gradient(model: tf.keras.Model, relu_policy: Callable) -> tf.keras.Model:
    """
    Given a model, commute all original ReLU by a new given ReLU policy.

    Parameters
    ----------
    model
        Model to commute.
    relu_policy
        Function wrapped with custom_gradient, defining the ReLU backprop behaviour.

    Returns
    -------
    model_commuted
    """
    cloned_model = clone_model(model)
    cloned_model.set_weights(model.get_weights())

    for layer_id in range(len(cloned_model.layers)): # pylint: disable=C0200
        layer = cloned_model.layers[layer_id]
        if has_relu_activation(layer):
            layer.activation = relu_policy()
        elif is_relu(layer):
            max_value = layer.max_value if hasattr(layer, 'max_value') else None
            threshold = layer.threshold if hasattr(layer, 'threshold') else None
            cloned_model.layers[layer_id].call = relu_policy(max_value, threshold)

    return cloned_model


def find_layer(model: tf.keras.Model, layer: Union[str, int]) -> tf.keras.layers.Layer:
    """
    Find a layer in a model either by his name or by his index.

    Parameters
    ----------
    model
        Model on which to search.
    layer
        Layer name or layer index

    Returns
    -------
    layer
        Layer found
    """
    if isinstance(layer, str):
        return model.get_layer(layer)
    if isinstance(layer, int):
        return model.layers[layer]
    raise ValueError(f"Could not find any layer {layer}.")



import numpy as np
import tensorflow as tf

from ..utils import almost_equal
from xplique.features_visualizations.transformations import random_blur, pad


def test_blur():
    """Test basic behaviour of the blur transformations"""
    imgs = np.zeros((3, 3, 3, 3), np.float32)
    # green dot
    imgs[:, 1, 1, 1] = 1.0
    # blue dot
    imgs[:, 2, 2, 2] = 1.0

    # assuming a gaussian kernel of size=3, sigma=1
    gaussian_kernel_sum = np.sum(np.exp(0)*1 + np.exp(-0.5)*4 + np.exp(-1)*4)
    c0 = np.exp(0)    / gaussian_kernel_sum # center
    c1 = np.exp(-0.5) / gaussian_kernel_sum # delta 1
    c2 = np.exp(-1.0) / gaussian_kernel_sum # delta 2

    blur_fn = random_blur(kernel_size=3, sigma_range=(1.0, 1.0))
    blur_imgs = blur_fn(imgs)

    # check we have the same images (batch blur is working correctly)
    assert almost_equal(blur_imgs[0],blur_imgs[1])
    assert almost_equal(blur_imgs[1],blur_imgs[2])
    # check that the channel are independant (not regular conv!)
    assert blur_imgs[0, 1, 1, 0] != blur_imgs[0, 1, 1, 1] and\
           blur_imgs[0, 1, 1, 1] != blur_imgs[0, 1, 1, 2]
    # assert the gaussian kernel is correct
    assert almost_equal(blur_imgs[0, :, :, 1], np.array([[c2, c1, c2],
                                                         [c1, c0, c1],
                                                         [c2, c1, c2]])) # green
    assert almost_equal(blur_imgs[0, :, :, 2], np.array([[0., 0., 0.],
                                                         [0., c2, c1],
                                                         [0., c1, c0]])) # blue
    assert almost_equal(blur_imgs[0, :, :, 0], np.array([[0., 0., 0.],
                                                         [0., 0., 0.],
                                                         [0., 0., 0.]])) # red


def test_pad():
    """Ensure the padding is consistent"""
    imgs = np.ones((3, 2, 2, 3), np.float32)
    pad_fn = pad(2, 0.0)
    imgs_padded = pad_fn(imgs)

    # check we have the same images (batch pad is working correctly)
    assert almost_equal(imgs_padded[0], imgs_padded[1])
    assert almost_equal(imgs_padded[1], imgs_padded[2])
    # check the channel are treated the same way
    assert almost_equal(imgs_padded[0, :, :, 0], imgs_padded[0, :, :, 1])
    assert almost_equal(imgs_padded[0, :, :, 1], imgs_padded[0, :, :, 2])
    # check the padding value is ok
    assert almost_equal(imgs_padded[0, :, :, 0], np.array(
                        [[0., 0., 0., 0., 0., 0.],
                         [0., 0., 0., 0., 0., 0.],
                         [0., 0., 1., 1., 0., 0.],
                         [0., 0., 1., 1., 0., 0.],
                         [0., 0., 0., 0., 0., 0.],
                         [0., 0., 0., 0., 0., 0.]]))



import numpy as np

from xplique.attributions import IntegratedGradients
from ..utils import generate_data, generate_model, almost_equal


def test_output_shape():
    """The output shape must be the same as the input shape"""

    input_shapes = [(28, 28, 1), (32, 32, 3)]
    nb_labels = 10

    for input_shape in input_shapes:
        x, y = generate_data(input_shape, nb_labels, 10)
        model = generate_model(input_shape, nb_labels)

        method = IntegratedGradients(model, -2, steps=100)
        outputs = method.explain(x, y)

        assert x.shape == outputs.shape


def test_straighline_path():
    """The path generated should be the straighline from the baseline to the input"""

    input_shapes = [(28, 28, 1), (32, 32, 3)]
    steps = 10
    x_value = 0.9

    for baseline_value in [0.0, 0.5, 1.0]:
        for input_shape in input_shapes:
            x = np.ones(input_shape, dtype=np.float32) * x_value
            baseline = IntegratedGradients._get_baseline(x.shape, baseline_value)
            path = IntegratedGradients._get_interpolated_points(x[None, :], steps, baseline[None, :])

            true_points = np.linspace(baseline_value, x_value, steps, dtype=np.float32)
            true_points = true_points[:, None, None, None] * np.ones(input_shape, dtype=np.float32)

            for point, true_point in zip(np.array(path), true_points):
                assert point.shape == x.shape
                assert point.min() == point.max()
                assert almost_equal(point, true_point)


def test_trapezoidal_rule():
    """The integral approximation should use trapezoidal rule"""

    # int( f(x) ) ~ sum( (f(a) + f(b)) / 2 ) for all points b > a
    points = np.array([0.1, 0.4, 0.6, 0.8, 0.9, 1.0, 0.8])

    true_trapezoidal = np.trapz(points, dx=1.0 / (len(points)-1))
    ig_trapezoidal = IntegratedGradients._average_gradients(points[None, :, None])

    assert almost_equal(ig_trapezoidal, true_trapezoidal)



"""
Module related to the Concept extraction mechanism
"""

import tensorflow as tf
import numpy as np

from sklearn.linear_model import SGDClassifier
from sklearn.svm import LinearSVC
from sklearn.model_selection import train_test_split

from ..commons import find_layer
from ..types import Union, Callable


class Cav: # pylint: disable=too-few-public-methods
    """
    Used to compute the Concept Activation Vector, which is a vector in the direction of the
    activations of that concept’s set of examples.

    Ref. Kim & al., Interpretability Beyond Feature Attribution: Quantitative Testing with Concept
    Activation Vectors (TCAV) (2018).
    https://arxiv.org/abs/1711.11279

    Parameters
    ----------
    model
        Model to extract concept from.
    target_layer
        Index of the target layer or name of the layer.
    classifier : 'SGD' or 'SVC' or Sklearn model, optional
        Default implementation use SGD with hinge classifier (linear SVM), SVC use libsvm but
        the computation time is longer.
    test_fraction
        Fraction of the dataset used for test
    batch_size
        Batch size during the activations extraction
    verbose
        If true, display information while training the classifier
    """

    def __init__(self,
                 model: tf.keras.Model,
                 target_layer: Union[str, int],
                 classifier: Union[str, Callable] = 'SGD',
                 test_fraction: float = 0.2,
                 batch_size: int = 64,
                 verbose: bool = False):
        self.model = model
        self.batch_size = batch_size
        self.test_fraction = test_fraction
        self.verbose = verbose

        # configure model bottleneck
        target_layer = find_layer(model, target_layer)
        self.bottleneck_model = tf.keras.Model(model.input, target_layer.output)

        # configure classifier
        if classifier == 'SGD':
            # official parameters
            self.classifier = SGDClassifier(alpha=0.01,
                                            max_iter=1_000,
                                            tol=1e-3,
                                            verbose=self.verbose)
        elif classifier == 'SVC':
            self.classifier = LinearSVC(verbose=self.verbose)
        elif all(hasattr(classifier, attr) for attr in ['fit', 'score']):
            self.classifier = classifier
        else:
            raise ValueError('The classifier passed is invalid.')

    def fit(self,
            positive_dataset: tf.Tensor,
            negative_dataset: tf.Tensor) -> tf.Tensor:
        """
        Compute and return the Concept Activation Vector (CAV) associated to the dataset and the
        layer targeted.

        Parameters
        ----------
        positive_dataset
            Dataset of positive samples : samples containing the concept.
        negative_dataset
            Dataset of negative samples : samples without the concept

        Returns
        -------
        cav
            Vector of the same shape as the layer output
        """
        positive_activations = self.bottleneck_model.predict(positive_dataset,
                                                             batch_size=self.batch_size)
        negative_activations = self.bottleneck_model.predict(negative_dataset,
                                                             batch_size=self.batch_size)

        activations = np.concatenate([positive_activations, negative_activations], axis=0)
        has_concept = np.array(
            [1 for _ in positive_activations] + [0 for _ in negative_activations], np.float32)

        original_shape = activations.shape
        activations = activations.reshape(len(activations), -1)

        x_train, x_test, y_train, y_test = train_test_split(activations, has_concept,
                                                            test_size=self.test_fraction)

        self.classifier.fit(x_train, y_train)

        if self.verbose:
            val_accuracy = self.classifier.score(x_test, y_test)
            print(f"[CAV] val_accuracy : {round(float(val_accuracy * 100), 2)}")

        # weights of each feature is the vector orthogonal to the hyperplane
        # note : x + epsilon * cav should increase the concept
        cav = self.classifier.coef_[0]
        cav = cav / np.linalg.norm(cav, 1)
        cav = np.array(cav).reshape(original_shape[1:])

        return tf.cast(cav, tf.float32)

    __call__ = fit



"""
Module related to Gradient x Input method
"""

import tensorflow as tf
import numpy as np

from .base import WhiteBoxExplainer, sanitize_input_output
from ..types import Optional, Union


class GradientInput(WhiteBoxExplainer):
    """
    Used to compute elementwise product between the saliency maps of Simonyan et al. and the
    input (Gradient x Input).

    Parameters
    ----------
    model
        The model from which we want to obtain explanations
    output_layer
        Layer to target for the outputs (e.g logits or after softmax).
        If an `int` is provided it will be interpreted as a layer index.
        If a `string` is provided it will look for the layer name.

        Default to the last layer.
        It is recommended to use the layer before Softmax.
    batch_size
        Number of inputs to explain at once, if None compute all at once.
    operator
        Function g to explain, g take 3 parameters (f, x, y) and should return a scalar,
        with f the model, x the inputs and y the targets. If None, use the standard
        operator g(f, x, y) = f(x)[y].
    """

    @sanitize_input_output
    def explain(self,
                inputs: Union[tf.data.Dataset, tf.Tensor, np.ndarray],
                targets: Optional[Union[tf.Tensor, np.ndarray]] = None) -> tf.Tensor:
        """
        Compute gradients x inputs for a batch of samples.

        Parameters
        ----------
        inputs
            Dataset, Tensor or Array. Input samples to be explained.
            If Dataset, targets should not be provided (included in Dataset).
            Expected shape among (N, W), (N, T, W), (N, H, W, C).
            More information in the documentation.
        targets
            Tensor or Array. One-hot encoding of the model's output from which an explanation
            is desired. One encoding per input and only one output at a time. Therefore,
            the expected shape is (N, output_size).
            More information in the documentation.

        Returns
        -------
        explanations
            Gradients x Inputs, with the same shape as the inputs.
        """
        gradients = self.batch_gradient(self.model, inputs, targets, self.batch_size)
        gradients_inputs = tf.multiply(gradients, inputs)

        return gradients_inputs



import numpy as np
import tensorflow as tf

from xplique.attributions import (Saliency, GradientInput, IntegratedGradients, SmoothGrad, VarGrad,
                                  SquareGrad, Occlusion, Rise, GuidedBackprop, DeconvNet, Lime,
                                  KernelShap)
from ..utils import generate_regression_model, generate_data

def _default_methods(model, output_layer_index):
    return [
        Saliency(model, output_layer_index),
        GradientInput(model, output_layer_index),
        SmoothGrad(model, output_layer_index),
        VarGrad(model, output_layer_index),
        SquareGrad(model, output_layer_index),
        IntegratedGradients(model, output_layer_index),
        GuidedBackprop(model, output_layer_index),
        DeconvNet(model, output_layer_index),
        Lime(model),
        KernelShap(model),
        Occlusion(model, patch_size=1, patch_stride=1),
    ]

def test_tabular_data():
    """Test applied to most attributions method"""

    features_shape, output_shape, samples = ((10,), 1, 20)
    model = generate_regression_model(features_shape, output_shape)
    output_layer_index = -1

    inputs_np, targets_np = generate_data(features_shape, output_shape, samples)
    inputs_tf, targets_tf = tf.cast(inputs_np, tf.float32), tf.cast(targets_np, tf.float32)
    dataset = tf.data.Dataset.from_tensor_slices((inputs_np, targets_np))
    # batched_dataset = tf.data.Dataset.from_tensor_slices((inputs_np, targets_np)).batch(4)

    methods = _default_methods(model, output_layer_index)

    for inputs, targets in [(inputs_np, targets_np),
                            (inputs_tf, targets_tf),
                            (dataset, None),
                            # (batched_dataset, None)
                            ]:
        for method in methods:
            try:
                explanations = method.explain(inputs, targets)
            except:
                raise AssertionError(
                    "Explanation failed for method ", method.__class__.__name__)

            # all explanation must have an explain method
            assert hasattr(method, 'explain')

            # all explanations returned must be numpy array
            assert isinstance(explanations, tf.Tensor)

            # all explanations shape should match features shape
            assert explanations.shape == [samples, *features_shape]

def test_multioutput_regression():
    """
    Test if in a multioutput regression settings the methods have the expected behavior
    """

    features_shape, output_shape, samples = ((10,), 4, 20)
    model = generate_regression_model(features_shape, output_shape=output_shape)
    output_layer_index = -1

    inputs_np, targets_np = generate_data(features_shape, output_shape, samples)
    inputs_tf, targets_tf = tf.cast(inputs_np, tf.float32), tf.cast(targets_np, tf.float32)
    dataset = tf.data.Dataset.from_tensor_slices((inputs_np, targets_np))
    # batched_dataset = tf.data.Dataset.from_tensor_slices((inputs_np, targets_np)).batch(4)

    methods = _default_methods(model, output_layer_index)

    for inputs, targets in [(inputs_np, targets_np),
                            (inputs_tf, targets_tf),
                            (dataset, None),
                            # (batched_dataset, None)
                            ]:
        for method in methods:
            try:
                explanations = method.explain(inputs, targets)
            except:
                raise AssertionError(
                    "Explanation failed for method ", method.__class__.__name__)

            # all explanation must have an explain method
            assert hasattr(method, 'explain')

            # all explanations returned must be numpy array
            assert isinstance(explanations, tf.Tensor)

            # all explanations shape should match features shape
            assert explanations.shape == [samples, *features_shape]

def test_batch_size():
    """
    Ensure the functioning of attributions for special batch size cases with tabular data
    """

    input_shape, nb_targets, samples = ((10,), 5, 20)
    inputs, targets = generate_data(input_shape, nb_targets, samples)
    model = generate_regression_model(input_shape, nb_targets)
    output_layer_index = -1

    batch_sizes = [None, 1, 32]

    for bs in batch_sizes:

        methods = [
            Saliency(model, output_layer_index, bs),
            GradientInput(model, output_layer_index, bs),
            SmoothGrad(model, output_layer_index, bs),
            VarGrad(model, output_layer_index, bs),
            SquareGrad(model, output_layer_index, bs),
            IntegratedGradients(model, output_layer_index, bs),
            GuidedBackprop(model, output_layer_index, bs),
            DeconvNet(model, output_layer_index, bs),
            Lime(model, bs),
            KernelShap(model, bs),
            Occlusion(model, bs, patch_size=1, patch_stride=1),
        ]

        for method in methods:
            try:
                explanations = method.explain(inputs, targets)
            except:
                raise AssertionError(
                    "Explanation failed for method ", method.__class__.__name__,
                    " batch size ", bs)



"""
Xplique
-------

The goal of Xplique is to provide a simple interface to the latest explanation
techniques
"""

__version__ = '1.1.0'

from . import attributions
from . import concepts
from . import features_visualizations
from . import commons
from . import plots



import tensorflow as tf

from xplique.attributions import SmoothGrad
from ..utils import generate_data, generate_model, almost_equal


def test_output_shape():
    """The output shape must be the same as the input shape"""

    input_shapes = [(28, 28, 1), (32, 32, 3)]
    nb_labels = 10

    for input_shape in input_shapes:
        x, y = generate_data(input_shape, nb_labels, 10)
        model = generate_model(input_shape, nb_labels)

        method = SmoothGrad(model, -2, batch_size=1, nb_samples=100)
        smoothed_gradients = method.explain(x, y)

        assert x.shape == smoothed_gradients.shape



import numpy as np

from xplique.commons import forgrad

from ..utils import generate_data, generate_model
from .test_common import _default_methods


def test_base_forgrad():
    """ Ensure forgrad is working with the expected shape """
    shapes = [ (5, 32, 32, 3), (5, 32, 32, 1), (5, 32, 32), (5, 31, 31), (5, 60, 60) ]
    sigmas = [5, 10, 30]

    for shape in shapes:
        heatmaps = np.random.rand(*shape).astype(np.float32)
        for sigma in sigmas:
            filtered_heatmaps = forgrad(heatmaps, sigma)
            assert filtered_heatmaps.shape[:3] == heatmaps.shape[:3]


def test_integration_forgrad():
    """ Test that forgrad integrate with all attributions """
    input_shape, nb_labels, samples = ((32, 32, 3), 10, 20)
    model = generate_model(input_shape, nb_labels)
    output_layer_index = -2

    inputs_np, targets_np = generate_data(input_shape, nb_labels, samples)
    methods = _default_methods(model, output_layer_index)

    for method in methods:
        explanations = method.explain(inputs_np, targets_np)
        explanations_filtered = forgrad(explanations)

        assert explanations_filtered.shape[:3] == explanations.shape[:3]





"""
Tests for Lime Module
"""
import numpy as np
import tensorflow as tf
from sklearn import linear_model
from xplique.attributions import Lime
from ..utils import generate_data, generate_model, almost_equal


def test_pertub_func():
    """Ensure the default pertub function has the right output"""
    num_features = 10
    pertub_func = Lime._get_default_pertub_function(prob = 0.5)
    samples = pertub_func(num_features, 12)

    assert samples.shape == (12,10)
    assert samples.dtype == tf.int32


def test_get_masks():
    """
    Ensure the get_masks function behave as expected (shape,type) and
    return correct values on toy examples
    """
    # say we have a mapping grouping an image by 2x2 pixels
    mapping_to_interpret_space = tf.constant([[0,0,1,1],[0,0,1,1],[2,2,3,3],[2,2,3,3]])
    # have four interpretation
    # first is the one representing the upper right of the image
    interpret_sample1 = [0,1,0,0]
    # second is the one representing the bottom left of the image
    interpret_sample2 = [0,0,1,0]
    # third contain the upper left and bottom right of the image
    interpret_sample3 = [1,0,0,1]
    # last one just miss the upper right
    interpret_sample4 = [1,0,1,1]

    interpret_samples = tf.constant([interpret_sample1,
                                     interpret_sample2,
                                     interpret_sample3,
                                     interpret_sample4])

    masks = Lime._get_masks(interpret_samples,mapping_to_interpret_space)
    assert masks.shape == (4,4,4)
    assert masks.dtype == tf.int32
    masks.numpy()
    expected_mask1 = np.array([[0,0,1,1],[0,0,1,1],[0,0,0,0],[0,0,0,0]])
    assert np.array_equal(masks[0],expected_mask1)

    expected_mask2 = np.array([[0,0,0,0],[0,0,0,0],[1,1,0,0],[1,1,0,0]])
    assert np.array_equal(masks[1],expected_mask2)

    expected_mask3 = np.array([[1,1,0,0],[1,1,0,0],[0,0,1,1],[0,0,1,1]])
    assert np.array_equal(masks[2],expected_mask3)

    expected_mask4 = np.array([[1,1,0,0],[1,1,0,0],[1,1,1,1],[1,1,1,1]])
    assert np.array_equal(masks[3],expected_mask4)

def test_apply_masks():
    """
    Ensure the apply_masks function behave as expected (shape,type) and
    work correctly on toy examples
    """
    # have a real artificial image to see if masks applies correctly
    img = tf.ones((2,2,3))
    # define 2 easy mask
    # we grayed all pixels excepts for the upper right pixel
    mask1 = [[0,1],[0,0]]
    # we grayed only the upper right pixel
    mask2 = [[1,0],[1,1]]
    # define the RGB value for pixel which will be masked
    baseline = tf.constant([125.0,255.0,0.0])

    masks = tf.constant([mask1,mask2])
    samples = Lime._apply_masks(img,masks,baseline)
    assert samples.shape == (2,2,2,3)
    assert samples.dtype == tf.float32

    samples.numpy()
    expected_sample1 = np.array([[[125,255,0],[1,1,1]],[[125,255,0],[125,255,0]]])
    assert np.array_equal(samples[0],expected_sample1)

    expected_sample2 = np.array([[[1,1,1],[125,255,0]],[[1,1,1],[1,1,1]]])
    assert np.array_equal(samples[1],expected_sample2)

def test_similarities():
    """
    Ensure that the compute similarity function behave as expected (shape, type)
    and return expected values on toy examples
    """
    original_input = tf.ones((2,2,3),dtype=tf.float32)
    mapping = tf.constant([[0,1],[2,3]], dtype=tf.int32)
    ref_value = tf.zeros(3, dtype=tf.float32)

    int_sample1 = np.array([1, 0, 0, 1])
    int_sample2 = np.array([0, 1, 0, 1])
    int_sample3 = np.array([0, 1, 0, 0])
    int_sample4 = np.array([0, 1, 1, 1])

    int_samples = tf.constant(
        [int_sample1, int_sample2, int_sample3, int_sample4],
        dtype=tf.int32
    )
    masks = Lime._get_masks(int_samples, mapping)
    pertubed_samples = Lime._apply_masks(original_input, masks, ref_value)

    similarity_kernel = Lime._get_exp_kernel_func()
    similarities = similarity_kernel(original_input, int_samples, pertubed_samples)

    assert similarities.shape == 4
    assert similarities.dtype == tf.float32

    expected_outcome = np.array([np.exp(-6),np.exp(-6),np.exp(-9),np.exp(-3)])

    similarities = similarities.numpy()
    assert almost_equal(similarities,expected_outcome)

    similarity_kernel2 = Lime._get_exp_kernel_func(distance_mode='cosine')
    similarities2 = similarity_kernel2(original_input, int_samples, pertubed_samples)

    assert similarities2.shape == 4
    assert similarities2.dtype == tf.float32

def test_compute():
    """The output shape must be the same as the input shape, except for the channels"""
    input_shapes = [(28, 28, 1), (32, 32, 3)]
    nb_labels = 10

    def map_four_by_four(inp):

        width = inp.shape[0]
        height = inp.shape[1]

        mapping = np.zeros((width,height))
        for i in range(width):
            if i%2 != 0:
                mapping[i] = mapping[i-1]
            else:
                for j in range(height):
                    mapping[i][j] = (width/2) * (i//2) + (j//2)

        mapping = tf.cast(mapping, dtype=tf.int32)
        return mapping

    for input_shape in input_shapes:
        samples, labels = generate_data(input_shape, nb_labels, 20)
        model = generate_model(input_shape, nb_labels)

        method = Lime(model,
                      interpretable_model=linear_model.Lasso(alpha=0.1),
                      map_to_interpret_space=map_four_by_four,
                      nb_samples=10,
                      kernel_width=10)

        explanations = method.explain(samples, labels)
        assert samples.shape[:3] == explanations.shape

def test_inputs_batching():
    """ Ensure, that we can call explain with batched inputs """
    nb_labels = 10

    samples, labels = generate_data((32, 32, 3), nb_labels, 200)
    model = generate_model((32, 32, 3), nb_labels)

    method = Lime(model,
                    batch_size=15,
                    interpretable_model=linear_model.Ridge(alpha=0.1),
                    nb_samples=20,
                    kernel_width=10)

    explanations = method.explain(samples, labels)
    assert samples.shape[:3] == explanations.shape



"""
Optimisation functions for MaCo method.
"""

import tensorflow as tf
import numpy as np

from ..types import Optional, Union, Callable, Tuple
from .preconditioning import maco_image_parametrization, init_maco_buffer
from .objectives import Objective


def maco(objective: Objective,
         optimizer: Optional[tf.keras.optimizers.Optimizer] = None,
         nb_steps: int = 256,
         noise_intensity: Optional[Union[float, Callable]] = 0.08,
         box_size: Optional[Union[float, Callable]] = None,
         nb_crops: Optional[int] = 32,
         values_range: Tuple[float, float] = (-1, 1),
         custom_shape: Optional[Tuple] = (512, 512)) -> Tuple[tf.Tensor, tf.Tensor]:
    """
    Optimise a single objective using MaCo method. Note that, unlike classic fourier optimization,
    we can only optimize for one objective at a time.

    Ref. Unlocking Feature Visualization for Deeper Networks with MAgnitude Constrained
         Optimization (2023).
    https://arxiv.org/abs/2306.06805

    Parameters
    ----------
    objective
        Objective object.
    optimizer
        Optimizer used for gradient ascent, default Nadam(lr=1.0).
    nb_steps
        Number of iterations.
    noise_intensity
        Control the noise injected at each step. Either a float : each step we add noise
        with same std, or a function that associate for each step a noise intensity.
    box_size
        Control the average size of the crop at each step. Either a fixed float (e.g 0.5 means
        the crops will be 50% of the image size) or a function that take as parameter the step
        and return the average box size. Default to linear decay from 50% to 5%.
    nb_crops
        Number of crops used at each steps, higher make the optimisation slower but
        make the results more stable. Default to 32.
    values_range
        Range of values of the inputs that will be provided to the model, e.g (0, 1) or (-1, 1).
    custom_shape
        If specified, optimizes images of the given size. Used with
        a low box size to optimize bigger images crop by crop.

    Returns
    -------
    image_optimized
        Optimized image for the given objective.
    transparency
        Transparency of the image, i.e the sum of the absolute value of the gradients
        of the image with respect to the objective.
    """
    values_range = (min(values_range), max(values_range))

    model, objective_function, _, input_shape = objective.compile()

    assert input_shape[0] == 1, "You can only optimize one objective at a time with MaCo."

    if optimizer is None:
        optimizer = tf.keras.optimizers.Nadam(1.0)

    if box_size is None:
        # default to box_size that go from 50% to 5%
        box_size_values = tf.cast(np.linspace(0.5, 0.05, nb_steps), tf.float32)
        get_box_size = lambda step_i: box_size_values[step_i]
    elif isinstance(box_size, Callable):
        get_box_size = box_size
    elif isinstance(box_size, float):
        get_box_size = lambda _ : box_size
    else:
        raise ValueError('box_size must be a function or a float.')

    if noise_intensity is None:
        # default to large noise to low noise
        noise_values = tf.cast(np.logspace(0, -4, nb_steps), tf.float32)
        get_noise_intensity = lambda step_i: noise_values[step_i]
    elif isinstance(noise_intensity, Callable):
        get_noise_intensity = noise_intensity
    elif isinstance(noise_intensity, float):
        get_noise_intensity = lambda _ : noise_intensity
    else:
        raise ValueError('noise_intensity size must be a function or a float.')

    img_shape = (input_shape[1], input_shape[2])
    if custom_shape:
        img_shape = custom_shape

    magnitude, phase = init_maco_buffer(img_shape)
    phase = tf.Variable(phase, trainable=True)

    transparency = tf.zeros((*custom_shape, 3))

    for step_i in range(nb_steps):

        box_size_at_i = get_box_size(step_i)
        noise_intensity_at_i = get_noise_intensity(step_i)

        grads, grads_img = maco_optimisation_step(model, objective_function, magnitude, phase,
                                                  box_size_at_i, noise_intensity_at_i, nb_crops,
                                                  values_range)

        optimizer.apply_gradients(zip([-grads], [phase]))
        transparency += tf.abs(grads_img)

    img = maco_image_parametrization(magnitude, phase, values_range).numpy()

    return img, transparency


@tf.function
def maco_optimisation_step(model, objective_function, magnitude, phase,
                           box_average_size, noise_std, nb_crops, values_range):
    """ Optimisation step for MaCo method.

    Parameters
    ----------
    model
        Model to optimize on.
    objective_function
        Objective function (e.g neurons, channel).
    magnitude
        Fixed magnitude used to generate the image (the magnitude is fixed and not optimized).
    phase
        Phase of the image of the current image in fourier domain.
    box_average_size
        Average size of the crops.
    noise_std
        Standard deviation of the noise added at each step.
    nb_crops
        Number of crops to use for each step.
    values_range
        Range of values of the inputs that will be provided to the model, e.g (0, 1) or (-1, 1).

    Returns
    -------
    grads
        Gradients of the phase with respect to the objective.
    grads_img
        Gradients of the image with respect to the objective.
    """
    with tf.GradientTape() as tape:
        tape.watch(phase)

        image = maco_image_parametrization(magnitude, phase, values_range)

        # sample random crops in the buffer
        center_x = 0.5 + tf.random.normal((nb_crops,), stddev=0.15)
        center_y = 0.5 + tf.random.normal((nb_crops,), stddev=0.15)
        delta_x = tf.random.normal((nb_crops,), stddev=0.05, mean=box_average_size)
        delta_x = tf.clip_by_value(delta_x, 0.05, 1.0)
        delta_y = delta_x  # square boxes

        box_indices = tf.zeros(shape=(nb_crops,), dtype=tf.int32)
        boxes = tf.stack([center_x - delta_x * 0.5,
                          center_y - delta_y * 0.5,
                          center_x + delta_x * 0.5,
                          center_y + delta_y * 0.5], -1)

        crops = tf.image.crop_and_resize(image[None, :, :, :], boxes, box_indices,
                                         model.input_shape[1:3])

        # add random noise as feature viz robustness
        # see (https://distill.pub/2017/feature-visualization)
        crops += tf.random.normal(crops.shape, stddev=noise_std, mean=0.0)
        crops += tf.random.uniform(crops.shape, minval=-noise_std/2.0,
                                                maxval=noise_std/2.0)

        # maco is always is a single objective
        model_outputs = model(crops)
        loss = tf.reduce_mean(objective_function([model_outputs]))

    # also get the gradient to the image for transparency
    grads = tape.gradient(loss, [phase, image])
    grads_phase, grads_image = grads

    return grads_phase, grads_image



"""
Ensure we can use the operator functionnality on various models
"""

import numpy as np
import tensorflow as tf

import pytest

from xplique.attributions import (Saliency, GradientInput, IntegratedGradients, SmoothGrad, VarGrad,
                                  SquareGrad, GradCAM, Occlusion, Rise, GuidedBackprop, DeconvNet,
                                  GradCAMPP, Lime, KernelShap, SobolAttributionMethod,
                                  HsicAttributionMethod)
from xplique.commons.operators import (check_operator, predictions_operator, regression_operator,
                                       binary_segmentation_operator, segmentation_operator)
from xplique.commons.operators import Tasks, get_operator
from xplique.commons.exceptions import InvalidOperatorException
from ..utils import generate_data, generate_regression_model


def default_methods(model, operator):
    return [
        Saliency(model, operator=operator),
        GradientInput(model, operator=operator),
        SmoothGrad(model, operator=operator),
        VarGrad(model, operator=operator),
        SquareGrad(model, operator=operator),
        IntegratedGradients(model, operator=operator),
        Occlusion(model, operator=operator),
        Rise(model, operator=operator, nb_samples=2),
        GuidedBackprop(model, operator=operator),
        DeconvNet(model, operator=operator),
        SobolAttributionMethod(model, operator=operator, grid_size=2, nb_design=2),
        HsicAttributionMethod(model, operator=operator, grid_size=2, nb_design=2),
    ]


def get_segmentation_model():
    model = tf.keras.Sequential([
        tf.keras.layers.Input((20, 20, 1)),
    ])
    model.compile()
    return model


def get_concept_model():
    model = tf.keras.Sequential([
        tf.keras.layers.Input((6)),
        tf.keras.layers.Dense((10))
    ])
    model.compile()
    return model


def test_check_operator():
    # ensure that the check operator detects non-operator

    # operator must have at least 3 arguments
    function_with_2_arguments = lambda x,y: 0

    # operator must be Callable
    not_a_function = [1, 2, 3]

    for operator in [function_with_2_arguments, not_a_function]:
        try:
            check_operator(operator)
            assert False
        except InvalidOperatorException:
            pass


def test_proposed_operators():
    # ensure all proposed operators are operators
    for operator in [predictions_operator, regression_operator,
                     binary_segmentation_operator, segmentation_operator]:
        check_operator(operator)


def test_get_operator():
    tasks_name = [task.name for task in Tasks]
    assert tasks_name.sort() == ['classification', 'regression'].sort()
    # get by enum
    assert get_operator(Tasks.CLASSIFICATION) is predictions_operator
    assert get_operator(Tasks.REGRESSION) is regression_operator

    # get by string
    assert get_operator("classification") is predictions_operator
    assert get_operator("regression") is regression_operator

    # assert a not valid string does not work
    with pytest.raises(AssertionError):
        get_operator("random")

    # operator must have at least 3 arguments
    function_with_2_arguments = lambda x,y: 0

    # operator must be Callable
    not_a_function = [1, 2, 3]

    for operator in [function_with_2_arguments, not_a_function]:
        try:
            get_operator(operator)
        except InvalidOperatorException:
            pass


def test_regression_operator():
    input_shape, nb_labels, samples = ((10, 10, 1), 10, 20)
    x, y = generate_data(input_shape, nb_labels, samples)
    regression_model = generate_regression_model(input_shape, nb_labels)

    methods = default_methods(regression_model, regression_operator)
    for method in methods:

        assert hasattr(method, 'inference_function')
        assert hasattr(method, 'batch_inference_function')
        assert hasattr(method, 'gradient')
        assert hasattr(method, 'batch_gradient')

        phis = method(x, y)

        assert x.shape[:-1] == phis.shape[:3]
    

def test_segmentation_operator():
    segmentation_model = get_segmentation_model()

    x, y = generate_data((20, 20, 3), 10, 10)

    def segmentation_operator(model, x, y):
        # explaining channel 0
        return tf.reduce_sum(model(x)[:,:,0], (1, 2))

    methods = default_methods(segmentation_model, segmentation_operator)
    for method in methods:

        assert hasattr(method, 'inference_function')
        assert hasattr(method, 'batch_inference_function')
        assert hasattr(method, 'gradient')
        assert hasattr(method, 'batch_gradient')

        phis = method(x, y)

        assert x.shape[:-1] == phis.shape[:3]


def test_concept_operator():
    concept_model = get_concept_model()

    x, y = generate_data((20, 20, 1), 10, 10)

    random_projection = tf.cast(np.random.uniform(size=(20*20, 6)), tf.float32)

    def concept_operator(model, x, y):
        x = tf.reshape(x, (-1, 20*20))
        print(x.shape, random_projection.shape)
        ui = x @ random_projection
        return tf.reduce_sum(model(ui) * y, axis=-1)

    methods = default_methods(concept_model, concept_operator)

    for method in methods:

        assert hasattr(method, 'inference_function')
        assert hasattr(method, 'batch_inference_function')
        assert hasattr(method, 'gradient')
        assert hasattr(method, 'batch_gradient')

        phis = method(x, y)

        assert x.shape[:-1] == phis.shape[:3]



"""
Numpy from/to Tensorflow manipulation
"""

import tensorflow as tf
import numpy as np

from ..types import Union, Optional, Tuple


def tensor_sanitize(inputs: Union[tf.data.Dataset, tf.Tensor, np.ndarray],
                    targets: Optional[Union[tf.Tensor, np.ndarray]]) -> Tuple[tf.Tensor, tf.Tensor]:
    """
    Ensure the output as tf.Tensor, accept various inputs format including:
    tf.Tensor, List, numpy array, tf.data.Dataset (when label = None).

    Parameters
    ----------
    inputs
        Input samples to be explained.
    targets
        One-hot encoded labels or regression target (e.g {+1, -1}), one for each sample.

    Returns
    -------
    inputs_tensors
        Inputs samples as tf.Tensor
    targets_tensors
        Targets as tf.Tensor
    """

    # deal with tf.data.Dataset
    if isinstance(inputs, tf.data.Dataset):
        # try to know if the dataset is batched, if it is the case we unbatch
        if hasattr(inputs, '_batch_size'):
            inputs = inputs.unbatch()
        # unpack the dataset, assume we have tuple of (input, target)
        targets = [target for _, target in inputs]
        inputs  = [inp for inp, _ in inputs]

    inputs = tf.cast(inputs, tf.float32)
    targets = tf.cast(targets, tf.float32)

    return inputs, targets


def numpy_sanitize(inputs: Union[tf.data.Dataset, tf.Tensor, np.ndarray],
                   targets: Optional[Union[tf.Tensor, np.ndarray]]) -> Tuple[tf.Tensor, tf.Tensor]:
    """
    Ensure the output as np.ndarray, accept various inputs format including:
    tf.Tensor, List, numpy array, tf.data.Dataset (when label = None).

    Parameters
    ----------
    inputs
        Input samples to be explained.
    targets
        One-hot encoded labels or regression target (e.g {+1, -1}), one for each sample.

    Returns
    -------
    inputs_ndarray
        Inputs samples as np.ndarray
    targets_ndarray
        Targets as np.ndarray
    """
    inputs, targets = tensor_sanitize(inputs, targets)
    return inputs.numpy(), targets.numpy()



"""
Custom callable operations
"""

import tensorflow as tf

from ..types import Callable

def predictions_one_hot_callable(
    model: Callable,
    inputs: tf.Tensor,
    targets: tf.Tensor) -> tf.Tensor:
    """
    Compute predictions scores, only for the label class, for a batch of samples.

    Parameters
    ----------
    model
        Model used for computing predictions.
    inputs
        Input samples to be explained.
    targets
        One-hot encoded labels or regression target (e.g {+1, -1}), one for each sample.

    Returns
    -------
    scores
        Predictions scores computed, only for the label class.
    """
    if isinstance(model, tf.lite.Interpreter):

        model.resize_tensor_input(0, [*inputs.shape], strict=False)
        model.allocate_tensors()
        model.set_tensor(model.get_input_details()[0]["index"], inputs)
        model.invoke()
        pred = model.get_tensor(model.get_output_details()[0]["index"])

    # can be a sklearn model or xgboost model
    elif hasattr(model, 'predict_proba'):
        pred = model.predict_proba(inputs.numpy())

    # can be another model thus it needs to implement a call function
    else:
        pred = model(inputs.numpy())

    # make sure that the prediction shape is coherent
    if inputs.shape[0] != 1:
        # a batch of prediction is required
        if len(pred.shape) == 1:
            # The prediction dimension disappeared
            pred = tf.expand_dims(pred, axis=1)

    pred = tf.cast(pred, dtype=tf.float32)
    scores = tf.reduce_sum(pred * targets, axis=-1)

    return scores



"""
Custom tensorflow operations
"""

import tensorflow as tf

from ..types import Optional, Union, Tuple


def repeat_labels(labels: tf.Tensor, nb_repetitions: int) -> tf.Tensor:
    """
    Duplicate each label nb_repetitions times.

    Parameters
    ----------
    labels
        One hot encoded labels (N, L) to compute for each sample, with N the number of samples,
        and L the number of classes.
    nb_repetitions
        Number of times each labels should be duplicate.

    Returns
    -------
    repeated_labels
        Unchanged label repeated (N*nb_repetitions, L).
    """
    repeated_labels = tf.expand_dims(labels, axis=1)
    repeated_labels = tf.repeat(repeated_labels, repeats=nb_repetitions, axis=1)

    repeated_labels = tf.reshape(repeated_labels, (-1, *repeated_labels.shape[2:]))

    return repeated_labels


def batch_tensor(tensors: Union[Tuple, tf.Tensor],
                 batch_size: Optional[int] = None):
    """
    Create a tensorflow dataset of tensors or series of tensors.

    Parameters
    ----------
    tensors
        Tuple of tensors or tensors to batch.
    batch_size
        Number of samples to iterate at once, if None process all at once.

    Returns
    -------
    dataset
        Tensorflow dataset batched.
    """
    dataset = tf.data.Dataset.from_tensor_slices(tensors)
    if batch_size is not None:
        dataset = dataset.batch(batch_size)

    return dataset
